
#!/usr/bin/env python
###############################################################################
# $Id$
#
# Project:  GDAL2Tiles, Google Summer of Code 2007 & 2008
#           Global Map Tiles Classes
# Purpose:  Convert a raster into TMS tiles, create KML SuperOverlay EPSG:4326,
#           generate a simple HTML viewers based on Google Maps and OpenLayers
# Author:   Klokan Petr Pridal, klokan at klokan dot cz
# Web:      http://www.klokan.cz/projects/gdal2tiles/
#
###############################################################################
# Copyright (c) 2008 Klokan Petr Pridal. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
###############################################################################

"""globalmaptiles.py
Global Map Tiles as defined in Tile Map Service (TMS) Profiles
==============================================================
Functions necessary for generation of global tiles used on the web.
It contains classes implementing coordinate conversions for:
  - GlobalMercator (based on EPSG:900913 = EPSG:3785)
       for Google Maps, Yahoo Maps, Microsoft Maps compatible tiles
  - GlobalGeodetic (based on EPSG:4326)
       for OpenLayers Base Map and Google Earth compatible tiles
More info at:
http://wiki.osgeo.org/wiki/Tile_Map_Service_Specification
http://wiki.osgeo.org/wiki/WMS_Tiling_Client_Recommendation
http://msdn.microsoft.com/en-us/library/bb259689.aspx
http://code.google.com/apis/maps/documentation/overlays.html#Google_Maps_Coordinates
Created by Klokan Petr Pridal on 2008-07-03.
Google Summer of Code 2008, project GDAL2Tiles for OSGEO.
In case you use this class in your product, translate it to another language
or find it usefull for your project please let me know.
My email: klokan at klokan dot cz.
I would like to know where it was used.
Class is available under the open-source GDAL license (www.gdal.org).
"""

import math

class GlobalMercator(object):
    """TMS Global Mercator Profile"""

    def __init__(self, tileSize=256):
        "Initialize the TMS Global Mercator pyramid"
        self.tileSize = tileSize
        self.initialResolution = 2 * math.pi * 6378137 / self.tileSize
        # 156543.03392804062 for tileSize 256 pixels
        self.originShift = 2 * math.pi * 6378137 / 2.0
        # 20037508.342789244

    def LatLonToMeters(self, lat, lon):
        "Converts given lat/lon in WGS84 Datum to XY in Spherical Mercator EPSG:900913"

        mx = lon * self.originShift / 180.0
        my = math.log( math.tan((90 + lat) * math.pi / 360.0 )) / (math.pi / 180.0)

        my = my * self.originShift / 180.0
        return mx, my

    def MetersToLatLon(self, mx, my):
        "Converts XY point from Spherical Mercator EPSG:900913 to lat/lon in WGS84 Datum"

        lon = (mx / self.originShift) * 180.0
        lat = (my / self.originShift) * 180.0

        lat = 180 / math.pi * (2 * math.atan( math.exp( lat * math.pi / 180.0)) - math.pi / 2.0)
        return lat, lon

    def PixelsToMeters(self, px, py, zoom):
        "Converts pixel coordinates in given zoom level of pyramid to EPSG:900913"

        res = self.Resolution( zoom )
        mx = px * res - self.originShift
        my = py * res - self.originShift
        return mx, my

    def MetersToPixels(self, mx, my, zoom):
        "Converts EPSG:900913 to pyramid pixel coordinates in given zoom level"
  
        res = self.Resolution( zoom )
        px = (mx + self.originShift) / res
        py = (my + self.originShift) / res
        return px, py

    def PixelsToTile(self, px, py):
        "Returns a tile covering region in given pixel coordinates"

        tx = int( math.ceil( px / float(self.tileSize) ) - 1 )
        ty = int( math.ceil( py / float(self.tileSize) ) - 1 )
        return tx, ty

    def PixelsToRaster(self, px, py, zoom):
        "Move the origin of pixel coordinates to top-left corner"

        mapSize = self.tileSize << zoom
        return px, mapSize - py

    def MetersToTile(self, mx, my, zoom):
        "Returns tile for given mercator coordinates"

        px, py = self.MetersToPixels(mx, my, zoom)
        return self.PixelsToTile(px, py)

    def TileBounds(self, tx, ty, zoom):
        "Returns bounds of the given tile in EPSG:900913 coordinates"

        minx, miny = self.PixelsToMeters( tx*self.tileSize, ty*self.tileSize, zoom )
        maxx, maxy = self.PixelsToMeters( (tx+1)*self.tileSize, (ty+1)*self.tileSize, zoom )
        return ( minx, miny, maxx, maxy )

    def TileLatLonBounds(self, tx, ty, zoom ):
        "Returns bounds of the given tile in latutude/longitude using WGS84 datum"

        bounds = self.TileBounds(tx, ty, zoom)
        minLat, minLon = self.MetersToLatLon(bounds[0], bounds[1])
        maxLat, maxLon = self.MetersToLatLon(bounds[2], bounds[3])

        return ( minLat, minLon, maxLat, maxLon )

    def Resolution(self, zoom):
        "Resolution (meters/pixel) for given zoom level (measured at Equator)"

        # return (2 * math.pi * 6378137) / (self.tileSize * 2**zoom)
        return self.initialResolution / (2**zoom)

    def ZoomForPixelSize(self, pixelSize):
        "Maximal scaledown zoom of the pyramid closest to the pixelSize."

        for i in range(30):
            if pixelSize > self.Resolution(i):
                return i-1 if i!=0 else 0 # We don't want to scale up

    def GoogleTile(self, tx, ty, zoom):
        "Converts TMS tile coordinates to Google Tile coordinates"

        # coordinate origin is moved from bottom-left to top-left corner of the extent
        return tx, (2**zoom - 1) - ty

    def GoogleToTMSTile(self, tx, ty, zoom):
        "Converts Google Tile coordinates to TMS tile coordinates"

        # coordinate origin is moved from top-left to bottom-left corner of the extent
        # The opposite of this: return tx, (2**zoom - 1) - ty
        return tx, (2**zoom - 1) - ty

    def QuadTree(self, tx, ty, zoom):
        "Converts TMS tile coordinates to Microsoft QuadTree"

        quadKey = ""
        ty = (2**zoom - 1) - ty
        for i in range(zoom, 0, -1):
            digit = 0
            mask = 1 << (i-1)
            if (tx & mask) != 0:
                digit += 1
            if (ty & mask) != 0:
                digit += 2
            quadKey += str(digit)

        return quadKey

#---------------------

class GlobalGeodetic(object):
    """TMS Global Geodetic Profile"""

    def __init__(self, tileSize = 256):
        self.tileSize = tileSize

    def LatLonToPixels(self, lat, lon, zoom):
        "Converts lat/lon to pixel coordinates in given zoom of the EPSG:4326 pyramid"

        res = 180 / 256.0 / 2**zoom
        px = (180 + lat) / res
        py = (90 + lon) / res
        return px, py

    def PixelsToTile(self, px, py):
        "Returns coordinates of the tile covering region in pixel coordinates"

        tx = int( math.ceil( px / float(self.tileSize) ) - 1 )
        ty = int( math.ceil( py / float(self.tileSize) ) - 1 )
        return tx, ty

    def Resolution(self, zoom ):
        "Resolution (arc/pixel) for given zoom level (measured at Equator)"

        return 180 / 256.0 / 2**zoom
        #return 180 / float( 1 << (8+zoom) )

    def TileBounds(self, tx, ty, zoom):
        "Returns bounds of the given tile"
        res = 180 / 256.0 / 2**zoom
        return (
            tx*256*res - 180,
            ty*256*res - 90,
            (tx+1)*256*res - 180,
            (ty+1)*256*res - 90
        )



"""### Calculate zonal statistics between a raster and a vector. ###

"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional

# External
import numpy as np
from osgeo import gdal, ogr
from numba import prange, jit

# Internal
from buteo.utils import (
    utils_path,
    utils_gdal,
    utils_projection,
)
from buteo.raster.core_raster import _get_basic_metadata_raster
from buteo.vector.core_vector import _get_basic_metadata_vector, vector_create_attribute_from_fid
from buteo.vector.rasterize import vector_rasterize
from buteo.raster.align import _raster_align_to_reference
from buteo.raster.core_raster_io import raster_to_array, array_to_raster
from buteo.raster.clip import raster_clip


# @jit(nopython=True, parallel=True, fastmath=True, nogil=True)
def _calculate_zonal_statistics(
    raster_arr: np.ndarray,
    vector_arr: np.ndarray,
    zones: np.ndarray,
    statistics: List[str],
) -> np.ndarray:
    """Internal. Calculate the zonal statistics."""
    # statistics
    stats = np.zeros((len(zones), len(statistics)), dtype=np.float32)

    # import pdb; pdb.set_trace()

    # calculate statistics
    for zone_idx in prange(len(zones)):
        zone = zones[zone_idx]
        if zone == 0:
            continue

        mask = vector_arr == zone
        for stat_idx, stat in enumerate(statistics):
            if stat == "mean":
                stats[zone_idx][stat_idx] = np.mean(raster_arr[mask])
            elif stat == "median":
                stats[zone_idx][stat_idx] = np.median(raster_arr[mask])
            elif stat == "std":
                stats[zone_idx][stat_idx] = np.std(raster_arr[mask])
            elif stat == "min":
                stats[zone_idx][stat_idx] = np.min(raster_arr[mask])
            elif stat == "max":
                stats[zone_idx][stat_idx] = np.max(raster_arr[mask])
            elif stat == "sum":
                stats[zone_idx][stat_idx] = np.sum(raster_arr[mask])

    return stats


def _raster_get_zonal_statistics(
    raster: Union[str, gdal.Dataset],
    vector: Union[str, gdal.Dataset],
    statistics: List[str] = ["mean"],
    out_path: Optional[str] = None,
) -> str:
    """TODO: Get zonal statistics between a raster and a vector."""
    assert isinstance(raster, (str, gdal.Dataset)), f"Invalid raster. {raster}"
    assert isinstance(vector, (str, ogr.DataSource)), f"Invalid vector. {vector}"
    assert isinstance(statistics, list), f"Invalid statistics. {statistics}"
    assert all([isinstance(stat, str) for stat in statistics]), f"Invalid statistics. {statistics}"

    if out_path is None:
        out_path = utils_path._get_temp_filepath("zonal_statistics.tif")

    raster_metadata = _get_basic_metadata_raster(raster)
    vector_metadata = _get_basic_metadata_vector(vector)

    # assert the projection is the same
    assert utils_projection._check_projections_match(raster_metadata["projection_osr"], vector_metadata["projection_osr"]), "Projections do not match."

    # clip raster to vector
    raster_clipped = raster_clip(raster, vector, add_timestamp=True, add_uuid=True)
    raster_clipped_metadata = _get_basic_metadata_raster(raster_clipped)

    vector_updated = vector_create_attribute_from_fid(vector, "_zonal_id")
    import pdb; pdb.set_trace()


    # rasterize vector
    rasterized_vector = vector_rasterize(
        vector,
        pixel_size=raster_clipped_metadata["pixel_size"],
        projection=raster_clipped_metadata["projection_osr"],
        extent=raster_clipped,
        out_path=out_path,
        burn_value=None,
        attribute="_zonal_id",
        dtype="int32",
    )

    rasterized_vector_aligned = _raster_align_to_reference(rasterized_vector, raster_clipped)[0]

    utils_gdal.delete_dataset_if_in_memory(rasterized_vector)

    # calculate zonal statistics
    rasterized_vector_arr = raster_to_array(rasterized_vector_aligned)
    raster_arr = raster_to_array(raster_clipped)

    utils_gdal.delete_dataset_if_in_memory(raster_clipped)
    utils_gdal.delete_dataset_if_in_memory(rasterized_vector_aligned)

    # zones
    zones = np.unique(rasterized_vector_arr)

    stats = _calculate_zonal_statistics(
        raster_arr,
        rasterized_vector_arr,
        zones,
        statistics,
    )



import numpy as np
from skimage import color, img_as_float, img_as_uint
from skimage.exposure import rescale_intensity
from skimage.util import view_as_blocks
from skimage.transform import resize as _resize
from scipy import signal, ndimage

"""
Adapted code from "Contrast Limited Adaptive Histogram Equalization" by Karel
Zuiderveld <karel@cv.ruu.nl>, Graphics Gems IV, Academic Press, 1994.

http://tog.acm.org/resources/GraphicsGems/gems.html#gemsvi

The Graphics Gems code is copyright-protected.  In other words, you cannot
claim the text of the code as your own and resell it. Using the code is
permitted in any program, product, or library, non-commercial or commercial.
Giving credit is not required, though is a nice gesture.  The code comes as-is,
and if there are any flaws or problems with any Gems code, nobody involved with
Gems - authors, editors, publishers, or webmasters - are to be held
responsible.  Basically, don't be a jerk, and remember that anything free
comes with no guarantee.
"""
def equalize_adapthist(image, ntiles_x=8, ntiles_y=8, clip_limit=0.01, nbins=256, nr_of_grey=16384):
    """Contrast Limited Adaptive Histogram Equalization.

    Parameters
    ----------
    image : array-like
        Input image.
    ntiles_x : int, optional
        Number of tile regions in the X direction.  Ranges between 2 and 16.
    ntiles_y : int, optional
        Number of tile regions in the Y direction.  Ranges between 2 and 16.
    clip_limit : float: optional
        Clipping limit, normalized between 0 and 1 (higher values give more
        contrast).
    nbins : int, optional
        Number of gray bins for histogram ("dynamic range").

    Returns
    -------
    out : ndarray
        Equalized image.

    Notes
    -----
    * The algorithm relies on an image whose rows and columns are even
      multiples of the number of tiles, so the extra rows and columns are left
      at their original values, thus  preserving the input image shape.
    * For color images, the following steps are performed:
       - The image is converted to LAB color space
       - The CLAHE algorithm is run on the L channel
       - The image is converted back to RGB space and returned
    * For RGBA images, the original alpha channel is removed.

    References
    ----------
    .. [1] http://tog.acm.org/resources/GraphicsGems/gems.html#gemsvi
    .. [2] https://en.wikipedia.org/wiki/CLAHE#CLAHE
    """
    args = [None, ntiles_x, ntiles_y, clip_limit * nbins, nbins]
    if image.ndim > 2:
        lab_img = color.rgb2lab(img_as_float(image))
        l_chan = lab_img[:, :, 0]
        l_chan /= np.max(np.abs(l_chan))
        l_chan = img_as_uint(l_chan)
        args[0] = rescale_intensity(l_chan, out_range=(0, nr_of_grey - 1))
        new_l = _clahe(*args).astype(float)
        new_l = rescale_intensity(new_l, out_range=(0, 100))
        lab_img[:new_l.shape[0], :new_l.shape[1], 0] = new_l
        image = color.lab2rgb(lab_img)
        image = rescale_intensity(image, out_range=(0, 1))

    else:
        image = img_as_uint(image)
        args[0] = rescale_intensity(image, out_range=(0, nr_of_grey - 1))
        out = _clahe(*args)
        image[:out.shape[0], :out.shape[1]] = out
        image = rescale_intensity(image)

    return image


def _clahe(image, ntiles_x, ntiles_y, clip_limit, nbins=128, nr_of_grey=16384, max_reg_x=16, max_reg_y=16):
    """Contrast Limited Adaptive Histogram Equalization.

    Parameters
    ----------
    image : array-like
        Input image.
    ntiles_x : int, optional
        Number of tile regions in the X direction.  Ranges between 2 and 16.
    ntiles_y : int, optional
        Number of tile regions in the Y direction.  Ranges between 2 and 16.
    clip_limit : float, optional
        Normalized clipping limit (higher values give more contrast).
    nbins : int, optional
        Number of gray bins for histogram ("dynamic range").
    nr_of_grey : int, optional
        Number of gray levels for image (2 ** bits_per_pixel).
    max_reg_x : int, optional
        Maximum number of contextual regions in the X direction.
    max_reg_y : int, optional
        Maximum number of contextual regions in the Y direction.

    Returns
    -------
    out : ndarray
        Equalized image.

    The number of "effective" greylevels in the output image is set by `nbins`;
    selecting a small value (eg. 128) speeds up processing and still produce
    an output image of good quality. The output image will have the same
    minimum and maximum value as the input image. A clip limit smaller than 1
    results in standard (non-contrast limited) AHE.
    """
    ntiles_x = min(ntiles_x, max_reg_x)
    ntiles_y = min(ntiles_y, max_reg_y)
    ntiles_y = max(ntiles_x, 2)
    ntiles_x = max(ntiles_y, 2)

    if clip_limit == 1.0:
        return image  # is OK, immediately returns original image.

    map_array = np.zeros((ntiles_y, ntiles_x, nbins), dtype=int)

    y_res = image.shape[0] - image.shape[0] % ntiles_y
    x_res = image.shape[1] - image.shape[1] % ntiles_x
    image = image[: y_res, : x_res]

    x_size = image.shape[1] // ntiles_x  # Actual size of contextual regions
    y_size = image.shape[0] // ntiles_y
    n_pixels = x_size * y_size

    if clip_limit > 0.0:  # Calculate actual cliplimit
        clip_limit = int(clip_limit * (x_size * y_size) / nbins)
        if clip_limit < 1:
            clip_limit = 1
    else:
        clip_limit = nr_of_grey  # Large value, do not clip (AHE)

    bin_size = 1 + nr_of_grey / nbins
    aLUT = np.arange(nr_of_grey)

    toto = 1.0 * aLUT
    toto //= bin_size
    aLUT = toto

    aLUT = np.round(aLUT)
    aLUT = np.abs(aLUT)
    aLUT = np.intp(aLUT)
    img_blocks = view_as_blocks(image, (y_size, x_size))

    # Calculate greylevel mappings for each contextual region
    for y in range(ntiles_y):
        for x in range(ntiles_x):
            sub_img = img_blocks[y, x]
            hist_indices = np.rint(sub_img.ravel())
            hist = aLUT[hist_indices.astype(int)] # pylint: disable=unsubscriptable-object
            hist = np.bincount(hist)
            hist = np.append(hist, np.zeros(nbins - hist.size, dtype=int))
            hist = clip_histogram(hist, clip_limit)
            hist = map_histogram(hist, 0, nr_of_grey - 1, n_pixels)
            map_array[y, x] = hist

    # Interpolate greylevel mappings to get CLAHE image
    ystart = 0
    for y in range(ntiles_y + 1):
        xstart = 0
        if y == 0:  # special case: top row
            ystep = y_size / 2.0
            yU = 0
            yB = 0
        elif y == ntiles_y:  # special case: bottom row
            ystep = y_size / 2.0
            yU = ntiles_y - 1
            yB = yU
        else:  # default values
            ystep = y_size
            yU = y - 1
            yB = yB + 1

        for x in range(ntiles_x + 1):
            if x == 0:  # special case: left column
                xstep = x_size / 2.0
                xL = 0
                xR = 0
            elif x == ntiles_x:  # special case: right column
                xstep = x_size / 2.0
                xL = ntiles_x - 1
                xR = xL
            else:  # default values
                xstep = x_size
                xL = x - 1
                xR = xL + 1

            mapLU = map_array[yU, xL]
            mapRU = map_array[yU, xR]
            mapLB = map_array[yB, xL]
            mapRB = map_array[yB, xR]

            xslice = np.arange(xstart, xstart + xstep)
            yslice = np.arange(ystart, ystart + ystep)
            interpolate(image, xslice, yslice, mapLU, mapRU, mapLB, mapRB, aLUT)

            xstart += xstep  # set pointer on next matrix */

        ystart += ystep

    return image


def clip_histogram(hist, clip_limit):
    """Perform clipping of the histogram and redistribution of bins.

    The histogram is clipped and the number of excess pixels is counted.
    Afterwards the excess pixels are equally redistributed across the
    whole histogram (providing the bin count is smaller than the cliplimit).

    Parameters
    ----------
    hist : ndarray
        Histogram array.
    clip_limit : int
        Maximum allowed bin count.

    Returns
    -------
    hist : ndarray
        Clipped histogram.
    """
    # calculate total number of excess pixels
    excess_mask = hist > clip_limit
    excess = hist[excess_mask]
    n_excess = excess.sum() - excess.size * clip_limit

    # Second part: clip histogram and redistribute excess pixels in each bin
    bin_incr = int(n_excess / hist.size)  # average binincrement
    upper = clip_limit - bin_incr  # Bins larger than upper set to cliplimit

    hist[excess_mask] = clip_limit

    low_mask = hist < upper
    n_excess -= hist[low_mask].size * bin_incr
    hist[low_mask] += bin_incr

    mid_mask = (hist >= upper) & (hist < clip_limit)
    mid = hist[mid_mask]
    n_excess -= mid.size * clip_limit - mid.sum()
    hist[mid_mask] = clip_limit

    while n_excess > 0:  # Redistribute remaining excess
        index = 0

        while n_excess > 0 and index < hist.size:
            step_size = int(hist[hist < clip_limit].size / n_excess)
            step_size = max(step_size, 1)
            indices = np.arange(index, hist.size, step_size)
            under = hist[indices] < clip_limit
            hist[under] += 1
            n_excess -= hist[under].size
            index += 1

    return hist


def map_histogram(hist, min_val, max_val, n_pixels):
    """Calculate the equalized lookup table (mapping).

    It does so by cumulating the input histogram.

    Parameters
    ----------
    hist : ndarray
        Clipped histogram.
    min_val : int
        Minimum value for mapping.
    max_val : int
        Maximum value for mapping.
    n_pixels : int
        Number of pixels in the region.

    Returns
    -------
    out : ndarray
       Mapped intensity LUT.
    """
    out = np.cumsum(hist).astype(float)
    scale = ((float)(max_val - min_val)) / n_pixels
    out *= scale
    out += min_val
    out[out > max_val] = max_val

    return out.astype(int)


def interpolate(image, xslice, yslice, mapLU, mapRU, mapLB, mapRB, aLUT):
    """Find the new grayscale level for a region using bilinear interpolation.

    Parameters
    ----------
    image : ndarray
        Full image.
    xslice, yslice : array-like
       Indices of the region.
    map* : ndarray
        Mappings of greylevels from histograms.
    aLUT : ndarray
        Maps grayscale levels in image to histogram levels.

    Returns
    -------
    out : ndarray
        Original image with the subregion replaced.

    Notes
    -----
    This function calculates the new greylevel assignments of pixels within
    a submatrix of the image. This is done by a bilinear interpolation between
    four different mappings in order to eliminate boundary artifacts.
    """
    norm = xslice.size * yslice.size  # Normalization factor

    # interpolation weight matrices
    x_coef, y_coef = np.meshgrid(np.arange(xslice.size),
                                 np.arange(yslice.size))
    x_inv_coef, y_inv_coef = x_coef[:, ::-1] + 1, y_coef[::-1] + 1

    view = image[int(yslice[0]):int(yslice[-1] + 1),
                 int(xslice[0]):int(xslice[-1] + 1)]
    view = np.rint(view).astype(int)

    im_slice = aLUT[view]
    new = ((y_inv_coef * (x_inv_coef * mapLU[im_slice]
                          + x_coef * mapRU[im_slice])
            + y_coef * (x_inv_coef * mapLB[im_slice]
                        + x_coef * mapRB[im_slice]))
           / norm)
    view[:, :] = new

    return image


def conv2SepMatlabbis(I, fen):
    rad = int((fen.size-1)/2)
    ligne = np.zeros((rad, I.shape[1]))
    I = np.append(ligne, I, axis=0)
    I = np.append(I, ligne, axis=0)

    colonne = np.zeros((I.shape[0], rad))
    I = np.append(colonne, I, axis=1)
    I = np.append(I, colonne, axis=1)

    res = conv2bis(conv2bis(I, fen.T), fen)

    return res


def EFolkiIter(I0, I1, iteration=5, radius=[8, 4], rank=4, uinit=None,vinit=None):
    talon=1.e-8
    if rank > 0:
        I0 = rank_filter_sup(I0, rank)
        I1 = rank_filter_sup(I1, rank)

    if uinit is None:
        u = np.zeros(I0.shape)
    else:
        u = uinit
    if vinit is None:
        v = np.zeros(I1.shape)
    else:
        v = vinit

    Iy, Ix = np.gradient(I0)

    cols, rows = I0.shape[1], I0.shape[0]
    x, y = np.meshgrid(range(cols), range(rows))

    for rad in radius:

        burt1D = np.array(np.ones([1, 2*rad+1]))/(2*rad + 1)
        W = lambda x: conv2SepMatlabbis(x, burt1D)

        Ixx = W(Ix*Ix) + talon
        Iyy = W(Iy*Iy) + talon
        Ixy = W(Ix*Iy)
        D = Ixx*Iyy - Ixy**2

        for _ in range(iteration):
            i1w = interp2(I1, x+u, y+v)

            it = I0 - i1w + u*Ix + v*Iy
            Ixt = W(Ix * it)
            Iyt = W(Iy * it)

            with np.errstate(divide='ignore', invalid='ignore'):
                u = (Iyy * Ixt - Ixy * Iyt) / D
                v = (Ixx * Iyt - Ixy * Ixt) / D

            invalid = np.isnan(u) | np.isinf(u) | np.isnan(v) | np.isinf(v)
            u[invalid] = 0
            v[invalid] = 0

    return u, v


def resize(image, shape, order=1):
    """Resize image to match a certain shape.

    Parameters
    ----------
    image : ndarray
        Input image.
    shape : tuple
        Shape of the output image.
    order : int, optional
        The order of the spline interpolation, default is 1.

    Returns
    -------
    out : ndarray
        Resized image.
    """
    return _resize(image, shape, order=order)


def GEFolkiIter(I0, I1, iteration=5, radius=[8, 4], rank=4, uinit=None, vinit=None):
    if rank > 0:
        R0 = rank_filter_sup(I0, rank)
        R1i = rank_filter_inf(I1, rank)
        R1s = rank_filter_sup(I1, rank)

    H0 = I0
    H1 = I1

    x = I0.shape[1]
    res_x = x % 8
    add_x = 8 - x % 8 if res_x > 0 else 0

    y = I0.shape[0]
    res_y = y % 8
    add_y = 8 - y % 8 if res_y > 0 else 0

    if res_x > 0 or res_y > 0:
        toto = resize(I0, (y+add_y, x+add_x), order=1)
    else:
        toto = I0

    toto = toto*255
    toto = toto.astype(np.uint8)
    H0 = equalize_adapthist(toto, 8, clip_limit=1, nbins=256)

    if res_x > 0 or res_y > 0:
        H0 = resize(H0, (y, x), order=1)

    H0 = H0.astype(np.float32)
    H0 = H0 / H0.max()

    x = I1.shape[1]
    res_x = x % 8
    add_x = 8 - x % 8 if res_x > 0 else 0

    y = I1.shape[0]
    res_y = y % 8
    add_y = 8 - y % 8 if res_y > 0 else 0

    if res_x > 0 or res_y > 0:
        toto = resize(I1, (y+add_y, x+add_x), order=1)
    else:
        toto = I1

    toto = toto * 255
    toto = toto.astype(np.uint8)
    H1 = equalize_adapthist(toto, 8, clip_limit=1, nbins=256)

    if res_x > 0 or res_y > 0:
        H1 = resize(H1, (y, x), order=1)

    H1 = H1.astype(np.float32)
    H1 = H1/H1.max()

    if uinit is None:
        u = np.zeros(I0.shape)
    else:
        u = uinit
    if vinit is None:
        v = np.zeros(I1.shape)
    else:
        v = vinit

    Iy, Ix = np.gradient(R0)

    cols, rows = I0.shape[1], I0.shape[0]
    x, y = np.meshgrid(range(cols), range(rows))
    for rad in radius:

        burt1D = np.array(np.ones([1, 2*rad+1]))/(2*rad + 1)
        W = lambda xin: conv2SepMatlabbis(xin, burt1D)

        Ixx = W(Ix*Ix)
        Iyy = W(Iy*Iy)
        Ixy = W(Ix*Iy)
        D = Ixx*Iyy - Ixy**2

        for _ in range(iteration):
            dx = x + u
            dy = y + v
            dx[dx < 0] = 0
            dy[dy < 0] = 0
            dx[dx > cols-1] = cols-1
            dy[dy > rows-1] = rows-1

            H1w = interp2(H1, dx, dy)

            crit1 = conv2SepMatlabbis(np.abs(H0-H1w), np.ones([2*rank+1, 1]))
            crit2 = conv2SepMatlabbis(np.abs(1-H0-H1w), np.ones([2*rank+1, 1]))

            R1w = interp2(R1s, x+u, y+v)
            R1w_1 = interp2(R1i, x+u, y+v)

            R1w[crit1 > crit2] = R1w_1[crit1 > crit2]
            it = R0 - R1w + u*Ix + v*Iy
            Ixt = W(Ix * it)
            Iyt = W(Iy * it)

            with np.errstate(divide='ignore', invalid='ignore'):
                u = (Iyy * Ixt - Ixy * Iyt) / D
                v = (Ixx * Iyt - Ixy * Ixt) / D

            invalid = np.isnan(u) | np.isinf(u) | np.isnan(v) | np.isinf(v)
            u[invalid] = 0
            v[invalid] = 0

    return u, v


def conv2(I, w):
    return signal.convolve2d(I, w, mode='valid')


def interp2(I, x, y, use_linear=True):
    if use_linear:
        return ndimage.map_coordinates(I, [y, x], order=1, mode='nearest').reshape(I.shape)

    return ndimage.map_coordinates(I, [y, x], order=3, mode='nearest')


def conv2bis(I, w):
    return signal.convolve2d(I, w, mode='valid')


def conv2Sep(I, w):
    return conv2(conv2(I, w.T), w)


class BurtOF:
    def __init__(self, flow, levels=4):
        self.flow = flow
        self.levels = levels

    def __call__(self, I0, I1, **kparams):

        if 'levels' in kparams:
            self.levels = kparams.pop('levels')

        I0 = (I0-I0.min())/(I0.max()-I0.min())
        I1 = (I1-I1.min())/(I1.max()-I1.min())

        Py0 = [I0]
        Py1 = [I1]

        for i in range(self.levels, 0, -1):
            Py0.append(self.pyrUp(Py0[-1]))
            Py1.append(self.pyrUp(Py1[-1]))

        u = np.zeros(Py0[-1].shape)
        v = np.zeros(Py0[-1].shape)

        for i in range(self.levels, -1, -1):
            kparams['uinit'] = u
            kparams['vinit'] = v
            u, v = self.flow(Py0[i], Py1[i], **kparams)
            if i > 0:
                col, row = Py0[i-1].shape[1], Py0[i-1].shape[0]
                u = 2 * self.pyrDown(u, (row, col))
                v = 2 * self.pyrDown(v, (row, col))

        return u, v

    def conv2SepMatlab(self, I, fen):
        rad = int((fen.size-1)/2)
        ligne = np.zeros((rad, I.shape[1]))
        I = np.append(ligne, I, axis=0)
        I = np.append(I, ligne, axis=0)

        colonne = np.zeros((I.shape[0], rad))
        I = np.append(colonne, I, axis=1)
        I = np.append(I, colonne, axis=1)

        res = conv2bis(conv2bis(I, fen.T), fen)

        return res

    def pyrUp(self, I):
        a = 0.4
        burt1D = np.array([[1./4.-a/2., 1./4., a, 1./4., 1./4.-a/2.]])

        M = self.conv2SepMatlab(I, burt1D)
        self.toto = M

        return M[::2, ::2]

    def pyrDown(self, I, shape):
        res = np.zeros(shape)
        I = np.repeat(np.repeat(I, 2, 0), 2, 1)
        col, row = I.shape[1], I.shape[0]
        col = min(shape[1], col)
        row = min(shape[0], row)
        res[:row, :col] = I[:row, :col]

        return res

def rank_filter_sup(I, rad):
    nl, nc = I.shape
    R = np.zeros([nl, nc])
    for i in range(-rad, rad+1):  # indice de ligne
        for j in range(-rad, rad+1):  # indice de colonne
            if i != 0:
                if i < 0:
                    tmp = np.concatenate([I[-i:, :], np.zeros([-i, nc])], axis=0)
                else:
                    tmp = np.concatenate([np.zeros([i, nc]), I[:-i, :]], axis=0)
            else:
                tmp = I
            if j != 0:
                if j < 0:
                    tmp = np.concatenate([tmp[:, -j:], np.zeros([nl, -j])], axis=1)
                else:
                    tmp = np.concatenate([np.zeros([nl, j]), tmp[:, :-j]], axis=1)

            idx = tmp > I
            R[idx] = R[idx]+1

    return R

def rank_filter_inf(I, rad):
    nl, nc = I.shape
    R = np.zeros([nl, nc])

    for i in range(-rad, rad+1):
        for j in range(-rad, rad+1):
            if i != 0:
                if i < 0:  # on decalle vers le haut de i lignes
                    tmp = np.concatenate([I[-i:, :], np.zeros([-i, nc])], axis=0)
                else:
                    tmp = np.concatenate([np.zeros([i, nc]), I[:-i, :]], axis=0)
            else:
                tmp = I
            if j != 0:
                if j < 0:
                    tmp = np.concatenate([tmp[:, -j:], np.zeros([nl, -j])], axis=1)
                else:
                    tmp = np.concatenate([np.zeros([nl, j]), tmp[:, :-j]], axis=1)

            idx = tmp < I
            R[idx] = R[idx]+1

    return R

def wrapData(I, u, v):
    """Apply the [u,v] optical flow to the data I"""
    col, row = I.shape[1], I.shape[0]
    X, Y = np.meshgrid(range(col), range(row))
    R = interp2(I, X+u, Y+v)

    return R



import sys; sys.path.append("../../")
from typing import Union, Optional, List

from osgeo import gdal
import numpy as np

from buteo.raster.core_raster_io import raster_to_array, array_to_raster
from buteo.raster.align import _raster_align_to_reference
from buteo.raster.reproject import _raster_reproject
from buteo.utils.utils_projection import _check_projections_match
from buteo.utils.utils_gdal import delete_dataset_if_in_memory


def coregister_images_efolki(
    master: Union[str, gdal.Dataset],
    slave: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    iteration: int = 4,
    radius: List[int] = [16, 8],
    rank: int = 4,
    levels: int = 5,
    band_to_base_master: int = 1,
    band_to_base_slave: int = 1,
    mask: Optional[np.ndarray] = None,
    fill_value: Optional[Union[int, float]] = None,
    resample_alg="nearest",
):
    """Coregister two images using the EFolki method. This method is based on the paper:

    "GeFolki: A Generic and Efficient Method for Optical Image Registration"

    Parameters
    ----------
    master : str
        Path to the master image.

    slave : str
        Path to the slave image.

    out_path : str
        Path to the output file. If None, the output will be written to a temporary file.

    iteration : int
        Number of iterations to run the algorithm.

    radius : list
        List of two integers specifying the radius of the search window.

    rank : int
        Rank of the algorithm.

    levels : int
        Number of levels to run the algorithm.

    band_to_base_master : int
        Band to use as the base for the coregistration.

    band_to_base_slave : int
        Band to use as the base for the coregistration.

    mask : np.ndarray
        Mask to apply to the coregistration.

    fill_value : int
        Fill value to use for the master and slave images.

    resample_alg : str
        The resampling algorithm, default: "nearest".

    Returns
    -------
    out_path : str
        Path to the output file.
    """
    try:
        from .gefolki import BurtOF, EFolkiIter, wrapData # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError("Using the GeFolki tools requires skimage and scipy. Please install them using `pip install scikit-image scipy`") from exc

    if not _check_projections_match(master, slave):
        reprojected = True
        reprojected_slave = _raster_reproject(slave, projection=master, resample_alg=resample_alg, copy_if_same=False)
    else:
        reprojected = False
        reprojected_slave = slave

    aligned_slave = _raster_align_to_reference(reprojected_slave, master, resample_alg=resample_alg)

    master_arr = raster_to_array(master, filled=True, fill_value=fill_value, cast=np.float32) # QB
    slave_arr = raster_to_array(aligned_slave, filled=True, fill_value=fill_value, cast=np.float32) # WV

    if reprojected:
        delete_dataset_if_in_memory(reprojected_slave)
        delete_dataset_if_in_memory(aligned_slave)

    if mask is not None:
        master_arr = master_arr * mask
        slave_arr = slave_arr * mask
    elif fill_value is not None:
        mask = np.logical_and(master_arr != fill_value, slave_arr != fill_value)
        master_arr = master_arr * mask
        slave_arr = slave_arr * mask

    EFolki = BurtOF(EFolkiIter)

    uu, vv = EFolki(
        slave_arr[:, :, band_to_base_slave - 1],
        master_arr[:, :, band_to_base_master - 1],
        iteration=iteration,
        radius=radius,
        rank=rank,
        levels=levels,
    )

    QBrecalee = wrapData(master_arr[:, :, band_to_base_master - 1], uu, vv)
    array_to_raster(QBrecalee[:, :, np.newaxis], reference=master, out_path=out_path)

    return out_path


def coregister_images_gefolki(
    master: Union[str, gdal.Dataset],
    slave: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    iteration: int = 4,
    radius: List[int] = [16, 8],
    rank: int = 4,
    levels: int = 5,
    band_to_base_master: int = 1,
    band_to_base_slave: int = 1,
    mask: Optional[np.ndarray] = None,
    fill_value: Optional[Union[int, float]] = None,
    resample_alg="nearest",
):
    """Coregister two images using the GeFolki method. This method is based on the paper:

    "GeFolki: A Generic and Efficient Method for Optical Image Registration"

    Parameters
    ----------
    master : str
        Path to the master image.

    slave : str
        Path to the slave image.

    out_path : str
        Path to the output file. If None, the output will be written to a temporary file.

    iteration : int
        Number of iterations to run the algorithm.

    radius : list
        List of two integers specifying the radius of the search window.

    rank : int
        Rank of the algorithm.

    levels : int
        Number of levels to run the algorithm.

    band_to_base_master : int
        Band to use as the base for the coregistration.

    band_to_base_slave : int
        Band to use as the base for the coregistration.

    mask : np.ndarray
        Mask to apply to the coregistration.

    fill_value : int
        Fill value to use for the master and slave images.

    Returns
    -------
    out_path : str
        Path to the output file.
    """
    try:
        from .gefolki import BurtOF, GEFolkiIter, wrapData # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError("Using the GeFolki tools requires skimage and scipy. Please install them using `pip install scikit-image scipy`") from exc

    if not _check_projections_match(master, slave):
        reprojected = True
        reprojected_slave = _raster_reproject(slave, projection=master, resample_alg=resample_alg, copy_if_same=False)
    else:
        reprojected = False
        reprojected_slave = slave

    aligned_slave = _raster_align_to_reference(reprojected_slave, master, resample_alg=resample_alg)

    master_arr = raster_to_array(master, filled=True, fill_value=fill_value, cast=np.float32) # QB
    slave_arr = raster_to_array(aligned_slave, filled=True, fill_value=fill_value, cast=np.float32) # WV

    if reprojected:
        delete_dataset_if_in_memory(reprojected_slave)
        delete_dataset_if_in_memory(aligned_slave)

    if mask is not None:
        master_arr = master_arr * mask
        slave_arr = slave_arr * mask
    elif fill_value is not None:
        mask = np.logical_and(master_arr != fill_value, slave_arr != fill_value)
        master_arr = master_arr * mask
        slave_arr = slave_arr * mask

    GEFolki = BurtOF(GEFolkiIter)

    uu, vv = GEFolki(
        slave_arr[:, :, band_to_base_slave - 1],
        master_arr[:, :, band_to_base_master - 1],
        iteration=iteration,
        radius=radius,
        rank=rank,
        levels=levels,
    )

    QBrecalee = wrapData(master_arr[:, :, band_to_base_master - 1], uu, vv)

    array_to_raster(QBrecalee[:, :, np.newaxis], reference=master, out_path=out_path)

    return out_path



"""Bindings to interact with the orfeo-toolbox in a more pythonic way.

TODO:
    - Documentation
"""

import sys; sys.path.append("../../") # Path: buteo/orfeo_toolbox_bindings.py
import os
import subprocess
from time import time

from buteo import raster_to_array


def _progress(percent, total, status=""):
    bar_len = 60
    filled_len = int(round(bar_len * percent / float(total)))

    percents = round(100.0 * percent / float(total), 1)
    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    print(f"[{bar}] {percents}% ... {status}", end="\r")
    if percents == 100:
        print("")
    sys.stdout.flush()


def execute_cli_function(command, name, quiet=False):
    process = subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    try:
        before = time()
        for line in iter(process.stdout.readline, ""):
            if "FATAL" in line:
                raise RuntimeError(line)
            elif "CRITICAL" in line:
                raise RuntimeError(line)
            elif "WARNING" in line:
                continue
            elif quiet is False:
                if "INFO" in line:
                    continue
            try:
                strip = line.strip()
                if len(strip) != 0:
                    part = strip.rsplit(":", 1)[1]
                    percent = int(part.split("%")[0])
                    _progress(percent, 100, name)
            except Exception:
                if len(line.strip()) != 0:
                    raise RuntimeError(line) from None

    except Exception:
        print("Critical failure while performing Orfeo-Toolbox action.")

    print(f"{name} completed in {round(time() - before, 2)}s.")


def otb_pansharpen(in_pan, in_xs, out_raster, options=None, out_datatype=None):
    """Pansharpen an image using the attributes
    of another image. Beware that the two images
    should be of the same size and position.
    """

    cli = "otbcli_Pansharpening"

    methods = ["rcs", "lmvm", "bayes"]

    if options is None:
        options = {
            "method": "lmvm",
            "method.lmvm.radiusx": 3,
            "method.lmvm.radiusy": 3,
        }

    if options["method"] not in methods:
        raise AttributeError("Selected method is not available.")

    if options["method"] == "lmvm":
        if "method.lmvm.radiusx" not in options:
            options["method.lmvm.radiusx"] = 3
        if "method.lmvm.radiusy" not in options:
            options["method.lmvm.radiusy"] = 3
    if options["method"] == "bayes":
        if "method.bayes.lamda" not in options:
            options["method.bayes.lamda"] = 0.9999
        if "method.bayes.s" not in options:
            options["method.bayes.s"] = 1

    if out_datatype is None:
        out_datatype = ""

    cli_args = [
        cli,
        "-inp",
        os.path.abspath(in_pan),
        "-inxs",
        os.path.abspath(in_xs),
        "-out",
        f'"{os.path.abspath(out_raster)}?&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
        out_datatype,
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="Pansharpening")

    return os.path.abspath(out_raster)


def otb_local_stats(in_raster, out_raster, options=None, band=None):
    """Computes local statistical moments on every pixel
    in the selected channel of the input image
    """

    cli = "otbcli_LocalStatisticExtraction"

    if options is None:
        options = {
            "channel": 1,
            "radius": 2,
        }

    if band is not None:
        band = f"&bands={band}"
    else:
        band = ""

    cli_args = [
        cli,
        "-in",
        os.path.abspath(in_raster),
        "-out",
        f'"{os.path.abspath(out_raster)}?{band}&gdal:co:COMPRESS=DEFLATE&gdal:co:PREDICTOR=3&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES" float',
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="Local statistics")

    return os.path.abspath(out_raster)


def otb_haralick(in_raster, out_raster, options=None, out_datatype="float", band=None):
    """Performs haralick texture extraction"""

    cli = "otbcli_HaralickTextureExtraction"

    stats_raster = raster_to_array(in_raster)
    stats = {"min": stats_raster.min(), "max": stats_raster.max()}
    stats_raster = None

    if options is None:
        options = {
            "texture": "simple",
            "channel": 1,
            "parameters.nbbin": 64,
            "parameters.xrad": 3,
            "parameters.yrad": 3,
            "parameters.min": stats["min"],
            "parameters.max": stats["max"],
        }

    if out_datatype is None:
        out_datatype = "float"

    if band is not None:
        band = f"&bands={band}"
    else:
        band = ""

    cli_args = [
        cli,
        "-in",
        os.path.abspath(in_raster),
        "-out",
        f'"{os.path.abspath(out_raster)}?{band}&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
        out_datatype,
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="Texture extraction")

    return os.path.abspath(out_raster)


def otb_dimension_reduction(in_raster, out_raster, options=None, out_datatype=None):
    """Performs dimensionality reduction on input image.
    PCA,NA-PCA,MAF,ICA methods are available.
    It is also possible to compute the inverse transform
    to reconstruct the image. It is also possible to
    optionally export the transformation matrix
    to a text file.
    """

    cli = "otbcli_DimensionalityReduction"

    if options is None:
        options = {
            "method": "pca",
            "rescale.outmin": 0,
            "rescale.outmax": 1,
            "nbcomp": 1,
            "normalize": "YES",
        }

    if out_datatype is None:
        out_datatype = ""

    cli_args = [
        cli,
        "-in",
        os.path.abspath(in_raster),
        "-out",
        f'"{os.path.abspath(out_raster)}?&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
        out_datatype,
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="Dimension reduction")

    return os.path.abspath(out_raster)


def otb_concatenate_images(in_rasters, out_raster, ram=None, out_datatype=None):
    """This application performs images channels concatenation.
    It reads the input image list (single or multi-channel) and
    generates a single multi-channel image. The channel order
    is the same as the list.
    """

    cli = "otbcli_ConcatenateImages"

    paths = []
    for raster in in_rasters:
        paths.append(os.path.abspath(raster))
    paths = " ".join(paths)

    if out_datatype is None:
        out_datatype = ""

    cli_string = " ".join(
        [
            cli,
            "-il",
            os.path.abspath(paths),
            "-out",
            f'"{os.path.abspath(out_raster)}?&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
            out_datatype,
            "-ram",
            str(ram),
        ]
    )

    execute_cli_function(cli_string, name="concatenate images")

    return os.path.abspath(out_raster)


def otb_split_images(in_raster, out_rasters, ram=None, out_datatype=None):
    """This application splits a N-bands image into N mono-band images.
    The output images filename will be generated from the output parameter.
    Thus, if the input image has 2 channels, and the user has set as
    output parameter, outimage.tif, the generated images will be
    outimage_0.tif and outimage_1.tif.
    """

    cli = "otbcli_SplitImage"

    if out_datatype is None:
        out_datatype = ""

    cli_string = " ".join(
        [
            cli,
            "-in",
            os.path.abspath(in_raster),
            "-out",
            f'"{os.path.abspath(out_rasters)}?&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
            out_datatype,
            "-ram",
            str(ram),
        ]
    )

    execute_cli_function(cli_string, name="splitting images")

    return out_rasters


def otb_rescale(in_raster, out_raster, options=None, out_datatype="float"):
    """This application scales the given image pixel intensity between two given values.
    By default min (resp. max) value is set to 0 (resp. 1).
    Input minimum and maximum values is automatically computed for all image bands.
    """

    cli = "otbcli_Rescale"

    if options is None:
        options = {
            "outmin": 0,
            "outmax": 1,
        }

    if out_datatype == "float":
        predictor = "gdal:co:PREDICTOR=3&"
    elif out_datatype == "uint16":
        predictor = "gdal:co:PREDICTOR=2&"
    else:
        predictor = ""

    cli_args = [
        cli,
        "-in",
        os.path.abspath(in_raster),
        "-out",
        f'"{os.path.abspath(out_raster)}?&gdal:co:COMPRESS=DEFLATE&{predictor}gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES"',
        out_datatype,
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="rescale image")

    return os.path.abspath(out_raster)


def otb_merge_rasters(
    in_rasters,
    out_raster,
    options=None,
    band=None,
    out_datatype="uint16",
    tmp=".",
    ram=42000,
    obt_path="C:/Program Files/OTB-7.4.0-Win64/",
    harmonisation=False,
    nodata_value=0,
    pixel_width=None,
    pixel_height=None,
):
    """Creates a mosaic out of a series of images. Must be of the same projection"""

    cli = "otbcli_Mosaic"

    if options is None:
        options = {
            "comp.feather": "slim",
            "comp.feather.slim.length": 1000,
            "harmo.method": "band",
            "harmo.cost": "rmse",
            "interpolator": "linear",
            "tmpdir": tmp,
            "nodata": nodata_value,
        }

        if pixel_width is not None and pixel_height is not None:
            options["output.spacingx"] = pixel_width
            options["output.spacingy"] = pixel_height

        if harmonisation:
            options["harmo.method"] = "band"
        else:
            options["harmo.method"] = "none"

    if band is not None:
        band = f"&bands={band}"
    else:
        band = ""

    cli_args = [
        cli,
        "-il",
        " ".join(in_rasters),
        "-out",
        f'"{os.path.abspath(out_raster)}?{band}&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES&gdal:co:TILED=YES"',
        out_datatype,
        "-ram",
        str(ram),
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    os.environ["PATH"] += os.pathsep + f"{obt_path}bin"
    os.environ["GDAL_DATA"] = os.pathsep + f"{obt_path}share/data"
    os.environ["PROJ_LIB"] = os.pathsep + f"{obt_path}share/proj"
    os.environ["OTB_APPLICATION_PATH"] = os.pathsep + f"{obt_path}lib/otb/applications"
    os.environ["PYTHONPATH"] = os.pathsep + f"{obt_path}lib/python"
    os.system(cli_string)

    execute_cli_function(cli_string, name="merge rasters")

    return os.path.abspath(out_raster)


# TODO: Try GeFolki for this
def otb_fine_registration(
    reference_image,
    secondary_image,
    warp_image,
    out_raster,
    out_raster_warp,
    options=None,
    band=None,
    ram=32000,
    otb_path="C:/Program Files/OTB-7.4.0-Win64/",
    erx=3,
    ery=3,
    mrx=3,
    mry=3,
    rgsx=3,
    rgsy=3,
    sgsx=3,
    sgsy=3,
    threads=-1,
):
    """Finely coregisters two images"""

    fn_name = "otbcli_FineRegistration"

    if options is None:
        options = {
            "erx": erx,
            "ery": ery,
            "mrx": mrx,
            "mry": mry,
            "rgsx": rgsx,
            "rgsy": rgsy,
            "sgsx ": sgsx,
            "sgsy": sgsy,
        }

    if band is not None:
        band = f"&bands={band}"
    else:
        band = ""

    cli_args = [
        fn_name,
        f'-ref "{os.path.abspath(reference_image)}"',
        f'-sec "{os.path.abspath(secondary_image)}"',
        f'-w "{os.path.abspath(warp_image)}"',
        "-out",
        f'"{os.path.abspath(out_raster)}?{band}&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES&gdal:co:TILED=YES"',
        "-wo",
        f'"{os.path.abspath(out_raster_warp)}?{band}&gdal:co:COMPRESS=DEFLATE&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES&gdal:co:TILED=YES"',
        "-ram",
        str(ram),
    ]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    if threads == -1:
        os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(os.cpu_count() * 2)
        os.environ["OPJ_NUM_THREADS"] = str(os.cpu_count() * 2)
        os.environ["GDAL_NUM_THREADS"] = str(os.cpu_count() * 2)
    else:
        os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(threads)
        os.environ["OPJ_NUM_THREADS"] = str(threads)
        os.environ["GDAL_NUM_THREADS"] = str(threads)

    os.environ["PATH"] += os.pathsep + os.path.join(os.path.abspath(otb_path), "bin")
    os.environ["GDAL_DATA"] = os.pathsep + os.path.join(os.path.abspath(otb_path), "share", "data")
    os.environ["PROJ_LIB"] = os.pathsep + os.path.join(os.path.abspath(otb_path), "share", "proj")
    os.environ["OTB_APPLICATION_PATH"] = os.pathsep + os.path.join(os.path.abspath(otb_path), "lib", "otb", "applications")
    os.environ["PYTHONPATH"] = os.pathsep + os.path.join(os.path.abspath(otb_path), "lib", "python")

    os.system(cli_string)

    return os.path.abspath(out_raster)


def obt_bandmath(
    in_rasters,
    expression,
    out_raster,
    band=None,
    ram=42000,
    out_datatype="float",
    obt_path="C:/Program Files/OTB-7.4.0-Win64/",
):
    """Creates a mosaic out of a series of images. Must be of the same projection"""

    cli = "otbcli_BandMath"

    if band is not None:
        band = f"&bands={band}"
    else:
        band = ""

    cli_args = [
        cli,
        "-il",
        " ".join(in_rasters),
        "-out",
        f'"{os.path.abspath(out_raster)}?{band}&gdal:co:COMPRESS=LZW&gdal:co:NUM_THREADS=ALL_CPUS&gdal:co:BIGTIFF=YES&gdal:co:TILED=YES"',
        out_datatype,
        "-ram",
        str(ram),
        f'-exp "{expression}"',
    ]

    cli_string = " ".join(cli_args)

    os.environ["PATH"] += os.pathsep + f"{obt_path}bin"
    os.environ["GDAL_DATA"] = os.pathsep + f"{obt_path}share/data"
    os.environ["PROJ_LIB"] = os.pathsep + f"{obt_path}share/proj"
    os.environ["OTB_APPLICATION_PATH"] = os.pathsep + f"{obt_path}lib/otb/applications"
    os.environ["PYTHONPATH"] = os.pathsep + f"{obt_path}lib/python"
    os.system(cli_string)

    execute_cli_function(cli_string, name="band_math")

    return os.path.abspath(out_raster)


def otb_meanshift_segmentation(
    in_raster,
    out_geom,
    spatialr=5,
    ranger=15,
    thres=0.1,
    maxiter=100,
    minsize=100,
    mask=False,
    stitch=True,
    neighbor=True,
    vector_minsize=1,
    tilesize=0,
):
    """Computes local statistical moments on every pixel
    in the selected channel of the input image
    """

    cli = "otbcli_Segmentation"

    options = {
        "mode": "vector",
        "mode.vector.out": out_geom,
        "mode.vector.neighbor": "false" if neighbor is False else "true",
        "mode.vector.stitch": "false" if stitch is False else "true",
        "mode.vector.tilesize": tilesize,
        "mode.vector.minsize": vector_minsize,
        "mode.vector.outmode": "ovw",
        "filter": "meanshift",
        "filter.meanshift.spatialr": spatialr,
        "filter.meanshift.ranger": ranger,
        "filter.meanshift.thres": thres,
        "filter.meanshift.maxiter": maxiter,
        "filter.meanshift.minsize": minsize,
    }

    if mask is not False:
        options["mode.vector.inmask"] = mask

    cli_args = [cli, "-in", os.path.abspath(in_raster)]

    for key, value in options.items():
        cli_args.append("-" + str(key))
        cli_args.append(str(value))

    cli_string = " ".join(cli_args)

    execute_cli_function(cli_string, name="meanshifting")


"""Create vector grids from references or hardcode"""
# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import ogr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
)
from buteo.vector import core_vector
from buteo.vector.metadata import _vector_to_metadata


# TODO: Implement grid functions
# TODO: create_grid, create_grid_points, create_grid_linestrings


def create_hexagonal_grid(extend, projection, cell_size, out_path=None):
    """Create a hexagonal grid in a given extend and projection.

    Args:
        extend (list): [xmin, ymin, xmax, ymax]
        projection (str): EPSG code
        cell_size (float): size of the hexagon
        out_path (str): path to save the grid

    Returns:
        None
    """
    # vector = core_vector._vector_open(extend)

    # vector.ExecuteSQL("CREATE TABLE hex_grid (id INTEGER PRIMARY KEY, geom GEOMETRY)")
    # vector.ExecuteSQL("SELECT AddGeometryColumn('hex_grid', 'geom', 4326, 'POLYGON', 2)")
    # vector.ExecuteSQL("CREATE INDEX hex_grid_geom_idx ON hex_grid USING GIST (geom)")
    # vector.ExecuteSQL("SELECT ST_Hex


"""Slope, aspect, hillshade, and other DEM functions."""

# Standard library
import os
import sys; sys.path.append("../../")
from typing import Union, Optional, List
from uuid import uuid4

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.raster.core_raster_io import raster_to_array, array_to_raster
from buteo.utils import utils_base, utils_gdal, utils_io, utils_path



def raster_dem_to_slope(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    out_path: Optional[Union[str, List[str]]] = None,
    slope_format: str = "percent",
    z_factor: float = 1.0,
    creation_options: Optional[List[str]] = None,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    out_format: str = "tif",
) -> str:
    """Calculate the slope of a raster DEM.

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be converted to slope.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    slope_format : str, optional
        "percent" or "degree", by default "percent".

    z_factor : float, optional
        Z factor for slope calculation, by default 1.0.

    creation_options : list, optional
        A list of GDAL creation options for the output raster(s).

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False

    out_format : str, optional
        The output format of the raster. If None, the format is inferred from the output path., default: ".tif"

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the raster(s) with slope.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(z_factor, [float, int], "z_factor")
    utils_base._type_check(slope_format, [str], "slope_format")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext=out_format,
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    slope_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        slopeFormat=slope_format,
        zFactor=z_factor,
        creationOptions=creation_options,
    )

    slope_rasters = []
    for idx, in_raster in enumerate(input_list):
        ds = gdal.DEMProcessing(
            output_list[idx],
            in_raster,
            'slope',
            options=slope_options,
        )

        if ds is None:
            raise RuntimeError(f"Error: Could not calculate slope: {in_raster}")

        ds.FlushCache()  # Write to disk.
        ds = None  # Close the dataset and release resources

        slope_rasters.append(output_list[idx])

    if input_is_list:
        return slope_rasters

    return slope_rasters[0]



def raster_dem_to_aspect(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    out_path: Optional[Union[str, List[str]]] = None,
    zero_for_flat: bool = True,
    creation_options: Optional[List[str]] = None,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    out_format: str = "tif",
) -> str:
    """Calculate the aspect in degrees of a raster DEM.

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be converted to aspect.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    zero_for_flat : bool, optional
        If True, set aspect to 0 for flat areas, by default False.

    creation_options : list, optional
        A list of GDAL creation options for the output raster(s).

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False

    out_format : str, optional
        The output format of the raster. If None, the format is inferred from the output path., default: ".tif"

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the raster(s) with slope.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(zero_for_flat, [bool], "zero_for_flat")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext=out_format,
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    aspect_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        creationOptions=creation_options,
        zeroForFlat=zero_for_flat,
    )

    aspect_rasters = []
    for idx, in_raster in enumerate(input_list):
        ds = gdal.DEMProcessing(
            output_list[idx],
            in_raster,
            'aspect',
            options=aspect_options,
        )

        if ds is None:
            raise RuntimeError(f"Error: Could not calculate slope: {in_raster}")

        ds.FlushCache()  # Write to disk.
        ds = None  # Close the dataset and release resources

        aspect_rasters.append(output_list[idx])

    if input_is_list:
        return aspect_rasters

    return aspect_rasters[0]



def raster_dem_to_hillshade(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    out_path: Optional[Union[str, List[str]]] = None,
    z_factor: float = 1.0,
    creation_options: Optional[List[str]] = None,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    out_format: str = "tif",
) -> str:
    """Calculate the hillshade of a raster DEM.

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be converted to hillshade.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    z_factor : float, optional
        Z factor for hillshade calculation, by default 1.0.

    creation_options : list, optional
        A list of GDAL creation options for the output raster(s).

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False

    out_format : str, optional
        The output format of the raster. If None, the format is inferred from the output path., default: ".tif"

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the raster(s) with slope.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(z_factor, [float, int], "z_factor")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext=out_format,
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    hillshade_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        creationOptions=creation_options,
        zFactor=z_factor,
    )

    hillshade_rasters = []
    for idx, in_raster in enumerate(input_list):
        ds = gdal.DEMProcessing(
            output_list[idx],
            in_raster,
            'hillshade',
            options=hillshade_options,
        )

        if ds is None:
            raise RuntimeError(f"Error: Could not calculate slope: {in_raster}")

        ds.FlushCache()  # Write to disk.
        ds = None  # Close the dataset and release resources

        hillshade_rasters.append(output_list[idx])

    if input_is_list:
        return hillshade_rasters

    return hillshade_rasters[0]


def raster_dem_to_orientation(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    out_path: Optional[Union[str, List[str]]] = None,
    include_height: bool = True,
    height_normalisation: bool = True,
    height_normalisation_value: float = 8849.0, # mt. everest
    creation_options: Optional[List[str]] = None,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    out_format: str = "tif",
) -> str:
    """Calculate the orientation of a raster DEM.

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be converted to orientation.

    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    include_height : bool, optional
        If True, include height in orientation calculation, by default False.

    height_normalisation : bool, optional
        If True, normalise height to 0-1, by default False.

    height_normalisation_value : float, optional
        Value to normalise height to, by default 8849m. The height of mt. everest.

    creation_options : list, optional
        A list of GDAL creation options for the output raster(s).

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False

    out_format : str, optional
        The output format of the raster. If None, the format is inferred from the output path., default: ".tif"

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the raster(s) with slope.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(include_height, [bool], "include_height")
    utils_base._type_check(height_normalisation, [bool], "height_normalisation")
    utils_base._type_check(height_normalisation_value, [float, int], "height_normalisation_value")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext=out_format,
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    creation_options = utils_gdal._get_default_creation_options(creation_options)

    orientation_rasters = []
    for idx, in_raster in enumerate(input_list):
        aspect_raster = raster_dem_to_aspect(in_raster, zero_for_flat=True)
        slope_raster = raster_dem_to_slope(in_raster, slope_format="percent")

        aspect = raster_to_array(aspect_raster)
        slope = raster_to_array(slope_raster)

        utils_gdal.delete_dataset_if_in_memory_list([
            aspect_raster,
            slope_raster,
        ])

        if include_height:
            dem = raster_to_array(in_raster)

            if height_normalisation:
                dem_norm = np.zeros(dem.shape, dtype=np.float32)
                dem_norm = np.divide(dem, height_normalisation_value, out=dem_norm, where=dem != 0)
            else:
                dem_norm = dem

            dem = None

        bands = 4 if include_height else 3

        destination = np.zeros((aspect.shape[0], aspect.shape[1], bands), dtype=np.float32)

        aspect_norm = np.zeros(aspect.shape, dtype=np.float32)
        aspect_norm = np.divide(aspect, 360.0, out=aspect_norm, where=aspect != 0)

        encoded_sin = ((np.sin(2 * np.pi * aspect_norm) + 1)) / 2.0
        encoded_cos = ((np.cos(2 * np.pi * aspect_norm) + 1)) / 2.0
        encoded_slope = np.clip(np.divide(slope, 90.0, where=slope != 0), 0.0, 1.0)

        destination[:, :, 0] = encoded_sin[:, :, 0]
        destination[:, :, 1] = encoded_cos[:, :, 0]
        destination[:, :, 2] = encoded_slope[:, :, 0]

        if include_height:
            destination[:, :, 3] = dem_norm[:, :, 0]

        array_to_raster(
            destination,
            reference=in_raster,
            out_path=output_list[idx],
            creation_options=creation_options,
        )

        orientation_rasters.append(output_list[idx])

    if input_is_list:
        return orientation_rasters

    return orientation_rasters[0]
